@@ -22,38 +22,40 @@
 
 %% @doc Basic file i/o operations for bitcask.
 -module(bitcask_fileops).
--author('Dave Smith <dizzyd@basho.com>').
--author('Justin Sheehy <justin@basho.com>').
--author('Andy Gross <andy@basho.com>').
 
 -export([create_file/2,
          open_file/1,
          close/1,
          close_for_writing/1,
+         data_file_tstamps/1,
          write/4,
          read/3,
          sync/1,
          delete/1,
          fold/3,
          fold_keys/3, fold_keys/4,
-         create_hintfile/1,
          mk_filename/2,
          filename/1,
          hintfile_name/1,
          file_tstamp/1,
-         tstamp/0,
          check_write/4]).
 
+-include_lib("kernel/include/file.hrl").
+
 -include("bitcask.hrl").
 
 -define(HINT_RECORD_SZ, 18). % Tstamp(4) + KeySz(2) + TotalSz(4) + Offset(8)
 
+-ifdef(PULSE).
+-compile({parse_transform, pulse_instrument}).
+-endif.
+
 -ifdef(TEST).
 -ifdef(EQC).
 -include_lib("eqc/include/eqc.hrl").
 -include_lib("eqc/include/eqc_fsm.hrl").
--compile(export_all).
 -endif.
+-compile(export_all).
 -include_lib("eunit/include/eunit.hrl").
 -endif.
 
@@ -61,13 +63,13 @@
 %% Called on a Dirname, will open a fresh file in that directory.
 -spec create_file(Dirname :: string(), Opts :: [any()]) -> {ok, #filestate{}}.
 create_file(DirName, Opts) ->
-    create_file_loop(DirName, Opts, tstamp()).
+    create_file_loop(DirName, [create] ++ Opts, most_recent_tstamp(DirName) + 1).
 
 %% @doc Open an existing file for reading.
 %% Called with fully-qualified filename.
 -spec open_file(Filename :: string()) -> {ok, #filestate{}} | {error, any()}.
 open_file(Filename) ->
-    case file:open(Filename, [read, raw, binary, read_ahead]) of
+    case bitcask_file:file_open(Filename, [readonly]) of
         {ok, FD} ->
             {ok, #filestate{mode = read_only,
                             filename = Filename, tstamp = file_tstamp(Filename),
@@ -80,26 +82,43 @@ open_file(Filename) ->
 -spec close(#filestate{} | fresh | undefined) -> ok.
 close(fresh) -> ok;
 close(undefined) -> ok;
-close(#filestate{ fd = FD, hintfd = HintFd }) ->
-    file:close(FD),
-    case HintFd of
-        undefined ->
-            ok;
-        _ ->
-            file:close(HintFd)
-    end,
+close(State = #filestate{ fd = FD }) ->
+    close_hintfile(State),
+    bitcask_file:file_close(FD),
     ok.
 
 %% @doc Close a file for writing, but leave it open for reads.
 -spec close_for_writing(#filestate{} | fresh | undefined) -> #filestate{} | ok.
 close_for_writing(fresh) -> ok;
 close_for_writing(undefined) -> ok;
-close_for_writing(State = 
-                  #filestate{ mode = read_write, fd = Fd, hintfd = HintFd }) ->
-    file:sync(Fd),
-    file:sync(HintFd),
-    file:close(HintFd),
-    State#filestate { mode = read_only, hintfd = undefined }.
+close_for_writing(State = #filestate{ mode = read_write, fd = Fd }) ->
+    S2 = close_hintfile(State),
+    bitcask_file:file_sync(Fd),
+    S2#filestate { mode = read_only }.
+
+close_hintfile(State = #filestate { hintfd = undefined }) ->
+    State;
+close_hintfile(State = #filestate { hintfd = HintFd, hintcrc = HintCRC }) ->
+    %% Write out CRC check at end of hint file.  Write with an empty key, zero
+    %% timestamp and offset as large as the file format supports so opening with
+    %% an older version of bitcask will just reject the record at the end of the
+    %% hintfile and otherwise work normally.
+    Iolist = hintfile_entry(<<>>, 0, {?MAXOFFSET, HintCRC}),
+    ok = bitcask_file:file_write(HintFd, Iolist),
+    bitcask_file:file_sync(HintFd),
+    bitcask_file:file_close(HintFd),
+    State#filestate { hintfd = undefined, hintcrc = 0 }.
+
+
+%% Build a list of {tstamp, filename} for all files in the directory that
+%% match our regex. 
+-spec data_file_tstamps(Dirname :: string()) -> [{integer(), string()}].
+data_file_tstamps(Dirname) ->
+    filelib:fold_files(Dirname, "^[0-9]+.bitcask.data$", false,
+                       fun(F, Acc) ->
+                               [{file_tstamp(F), F} | Acc]
+                       end, []).
+
 
 %% @doc Use only after merging, to permanently delete a data file.
 -spec delete(#filestate{}) -> ok | {error, atom()}.
@@ -119,7 +138,8 @@ delete(#filestate{ filename = FN } = State) ->
         {error, read_only}.
 write(#filestate { mode = read_only }, _K, _V, _Tstamp) ->
     {error, read_only};
-write(Filestate=#filestate{fd = FD, hintfd = HintFD, ofs = Offset},
+write(Filestate=#filestate{fd = FD, hintfd = HintFD, 
+                           hintcrc = HintCRC0, ofs = Offset},
       Key, Value, Tstamp) ->
     KeySz = size(Key),
     true = (KeySz =< ?MAXKEYSIZE),
@@ -131,14 +151,16 @@ write(Filestate=#filestate{fd = FD, hintfd = HintFD, ofs = Offset},
               <<ValueSz:?VALSIZEFIELD>>, Key, Value],
     Bytes  = [<<(erlang:crc32(Bytes0)):?CRCSIZEFIELD>> | Bytes0],
     %% Store the full entry in the data file
-    ok = file:pwrite(FD, Offset, Bytes),
+    ok = bitcask_file:file_pwrite(FD, Offset, Bytes),
     %% Create and store the corresponding hint entry
     TotalSz = KeySz + ValueSz + ?HEADER_SIZE,
     Iolist = hintfile_entry(Key, Tstamp, {Offset, TotalSz}),
-    ok = file:write(HintFD, Iolist),
+    ok = bitcask_file:file_write(HintFD, Iolist),
     %% Record our final offset
-    FinalSz = iolist_size(Bytes),
-    {ok, Filestate#filestate{ofs = Offset + FinalSz}, Offset, FinalSz}.
+    TotalSz = iolist_size(Bytes),
+    HintCRC = erlang:crc32(HintCRC0, Iolist), % compute crc of hint
+    {ok, Filestate#filestate{ofs = Offset + TotalSz,
+                            hintcrc = HintCRC}, Offset, TotalSz}.
 
 
 %% @doc Given an Offset and Size, get the corresponding k/v from Filename.
@@ -154,19 +176,21 @@ read(Filename, Offset, Size) when is_list(Filename) ->
             {error, Reason}
     end;
 read(#filestate { fd = FD }, Offset, Size) ->
-    case file:pread(FD, Offset, Size) of
+    case bitcask_file:file_pread(FD, Offset, Size) of
         {ok, <<Crc32:?CRCSIZEFIELD/unsigned, Bytes/binary>>} ->
-            %% Unpack the actual data
-            <<_Tstamp:?TSTAMPFIELD, KeySz:?KEYSIZEFIELD, ValueSz:?VALSIZEFIELD,
-             Key:KeySz/bytes, Value:ValueSz/bytes>> = Bytes,
-
             %% Verify the CRC of the data
             case erlang:crc32(Bytes) of
                 Crc32 ->
+                    %% Unpack the actual data
+                    <<_Tstamp:?TSTAMPFIELD,
+                     KeySz:?KEYSIZEFIELD, ValueSz:?VALSIZEFIELD,
+                     Key:KeySz/bytes, Value:ValueSz/bytes>> = Bytes,
                     {ok, Key, Value};
                 _BadCrc ->
                     {error, bad_crc}
             end;
+        eof ->
+            {error, eof};
         {error, Reason} ->
             {error, Reason}
     end.
@@ -174,8 +198,8 @@ read(#filestate { fd = FD }, Offset, Size) ->
 %% @doc Call the OS's fsync(2) system call on the cask and hint files.
 -spec sync(#filestate{}) -> ok.
 sync(#filestate { mode = read_write, fd = Fd, hintfd = HintFd }) ->
-    ok = file:sync(Fd),
-    ok = file:sync(HintFd).
+    ok = bitcask_file:file_sync(Fd),
+    ok = bitcask_file:file_sync(HintFd).
 
 -spec fold(fresh | #filestate{},
            fun((binary(), binary(), integer(),
@@ -185,15 +209,25 @@ sync(#filestate { mode = read_write, fd = Fd, hintfd = HintFd }) ->
 fold(fresh, _Fun, Acc) -> Acc;
 fold(#filestate { fd=Fd, filename=Filename, tstamp=FTStamp }, Fun, Acc) ->
     %% TODO: Add some sort of check that this is a read-only file
-    {ok, _} = file:position(Fd, bof),
-    case file:read(Fd, ?HEADER_SIZE) of
+    %% TODO: Need to position+read?!
+    ok = bitcask_file:file_seekbof(Fd),
+    case bitcask_file:file_read(Fd, ?HEADER_SIZE) of
         {ok, <<_Crc:?CRCSIZEFIELD, _Tstamp:?TSTAMPFIELD, _KeySz:?KEYSIZEFIELD,
               _ValueSz:?VALSIZEFIELD>> = H} ->
-            fold_loop(Fd, Filename, FTStamp, H, 0, Fun, Acc);
+            fold_loop(Fd, Filename, FTStamp, H, 0, Fun, Acc, 0);
+        {ok, OtherBytes} ->
+            error_logger:error_msg("~s:fold: ~s: expected ~p bytes but got "
+                                   "only ~p bytes, skipping\n",
+                                   [?MODULE, Filename, ?HEADER_SIZE,
+                                    size(OtherBytes)]),
+            Acc;
         eof ->
             Acc;
         {error, Reason} ->
-            {error, Reason}
+            %% A truncated file would yield {ok, OtherBytes} or eof.
+            %% Something really bad has happened, either an allocation
+            %% error or something Truly Bad, e.g. EBADF, EIO.
+            throw({fold_error, {read_read, Filename, 0, ?HEADER_SIZE, Reason}, Acc})
     end.
 
 -spec fold_keys(fresh | #filestate{}, fun((binary(), integer(), {integer(), integer()}, any()) -> any()), any()) ->
@@ -202,7 +236,7 @@ fold_keys(fresh, _Fun, Acc) -> Acc;
 fold_keys(State, Fun, Acc) ->
     fold_keys(State, Fun, Acc, default).
 
--spec fold_keys(fresh | #filestate{}, fun((binary(), integer(), {integer(), integer()}, any()) -> any()), any(), datafile | hintfile | default) ->
+-spec fold_keys(fresh | #filestate{}, fun((binary(), integer(), {integer(), integer()}, any()) -> any()), any(), datafile | hintfile | default | recovery) ->
         any() | {error, any()}.
 fold_keys(#filestate { fd = Fd } = State, Fun, Acc, Mode) ->
     case Mode of
@@ -216,35 +250,23 @@ fold_keys(#filestate { fd = Fd } = State, Fun, Acc, Mode) ->
                     fold_hintfile(State, Fun, Acc);
                 false ->
                     fold_keys_loop(Fd, 0, Fun, Acc)
+            end;
+        recovery -> % if hint files are corrupt, restart scanning cask files
+                    % Fun should be side-effect free or tolerant of being
+                    % called twice
+            case has_valid_hintfile(State) of
+                true ->
+                    case fold_hintfile(State, Fun, Acc) of
+                        {error, _} ->
+                            fold_keys_loop(Fd, 0, Fun, Acc);
+                        Acc1 ->
+                            Acc1
+                    end;
+                false ->
+                    fold_keys_loop(Fd, 0, Fun, Acc)
             end
     end.
 
--spec create_hintfile(string() | #filestate{}) -> ok | {error, any()}.
-create_hintfile(Filename) when is_list(Filename) ->
-    case open_file(Filename) of
-        {ok, Fstate} ->
-            try
-                create_hintfile(Fstate)
-            after
-                close(Fstate)
-            end;
-        {error, Reason} ->
-            {error, Reason}
-    end;
-create_hintfile(State) when is_record(State, filestate) ->
-    F = fun(K, Tstamp, {Offset, TotalSz}, HintFd) ->
-                Iolist = hintfile_entry(K, Tstamp, {Offset, TotalSz}),
-                case file:write(HintFd, Iolist) of
-                    ok ->
-                        HintFd;
-                    {error, Reason} ->
-                        throw({error, Reason})
-                end
-        end,
-    generate_hintfile(hintfile_name(State),
-                      {?MODULE, fold_keys_loop, [State#filestate.fd, 0, F]}).
-
-
 -spec mk_filename(string(), integer()) -> string().
 mk_filename(Dirname, Tstamp) ->
     filename:join(Dirname,
@@ -283,48 +305,80 @@ check_write(#filestate { ofs = Offset }, Key, Value, MaxSize) ->
 has_hintfile(#filestate { filename = Fname }) ->
     filelib:is_file(hintfile_name(Fname)).
 
+%% Return true if there is a hintfile and it has
+%% a valid CRC check
+has_valid_hintfile(State) ->
+    case has_hintfile(State) of
+        true ->
+            case fold_hintfile(State, fun has_valid_hintfile_visitor/4, 0) of
+                N when is_number(N) ->
+                    true;
+                _ ->
+                    false
+            end;
+        X ->
+            X
+    end.
+
+
 
 %% ===================================================================
 %% Internal functions
 %% ===================================================================
 
-%% @private
-tstamp() ->
-    {Mega, Sec, _Micro} = now(),
-    (Mega * 1000000) + Sec.
-
-
-fold_loop(Fd, Filename, FTStamp, Header, Offset, Fun, Acc0) ->
+fold_loop(_Fd, Filename, _FTStamp, _Header, Offset, _Fun, Acc, 20) ->
+    error_logger:error_msg("fold_loop: CRC error limit at file ~p offset ~p\n",
+                           [Filename, Offset]),
+    Acc;
+fold_loop(Fd, Filename, FTStamp, Header, Offset, Fun, Acc0, CrcSkipCount) ->
     <<Crc32:?CRCSIZEFIELD, Tstamp:?TSTAMPFIELD, KeySz:?KEYSIZEFIELD,
      ValueSz:?VALSIZEFIELD>> = Header,
     <<_:4/binary, HeaderMinusCRC/binary>> = Header,
     TotalSz = KeySz + ValueSz + ?HEADER_SIZE,
-    case file:read(Fd, TotalSz) of
+    case bitcask_file:file_read(Fd, TotalSz) of
         {ok, <<Key:KeySz/bytes, Value:ValueSz/bytes, Rest/binary>>} ->
-            case erlang:crc32([HeaderMinusCRC, Key, Value]) of
-                Crc32 ->
-                    PosInfo = {Filename, FTStamp, Offset, TotalSz},
-                    Acc = Fun(Key, Value, Tstamp, PosInfo, Acc0),
-                    case Rest of
-                        <<NextHeader:?HEADER_SIZE/bytes>> ->
-                            fold_loop(Fd, Filename, FTStamp, NextHeader,
-                                      Offset + TotalSz, Fun, Acc);
-                        <<>> ->
-                            Acc
-                    end;
-                _ ->
-                    {error, {bad_crc, Fd, Offset}}
+            CrcMatch = erlang:crc32([HeaderMinusCRC, Key, Value]) =:= Crc32,
+            Acc = case CrcMatch of
+                      true ->
+                          PosInfo = {Filename, FTStamp, Offset, TotalSz},
+                          Fun(Key, Value, Tstamp, PosInfo, Acc0);
+                      false ->
+                          error_logger:error_msg(
+                            "fold_loop: CRC error at file ~s offset ~p, "
+                            "skipping ~p bytes\n", [Filename, Offset, TotalSz]),
+                          Acc0
+                  end,
+            case Rest of
+                <<NextHeader:?HEADER_SIZE/bytes>> ->
+                    NewCrcSkipCount = if (not CrcMatch) -> CrcSkipCount + 1;
+                                         true           -> CrcSkipCount
+                                      end,
+                    fold_loop(Fd, Filename, FTStamp, NextHeader,
+                              Offset + TotalSz, Fun, Acc, NewCrcSkipCount);
+                <<>> ->
+                    Acc;
+                Tail ->
+                    error_logger:error_msg(
+                      "Trailing data, discarding (~p bytes)\n",
+                      [size(Tail)]),
+                    Acc
             end;
         {ok, X} ->
             error_logger:error_msg("Bad datafile entry, discarding"
                                    "(~p/~p bytes)\n", [size(X),TotalSz]),
             Acc0;
+        eof ->
+            error_logger:error_msg("Unexpected EOF, ignore (~p bytes)\n",
+                [TotalSz]),
+            Acc0;
         {error, Reason} ->
-            {error, Reason}
+            %% Again, either we had an allocation error in NIF land,
+            %% or something Truly Bad happened, e.g. EBADF, EIO
+            throw({fold_error, {file_read, Filename, Offset, TotalSz, Reason}, Acc0})
     end.
 
 fold_keys_loop(Fd, Offset, Fun, Acc0) ->
-    case file:pread(Fd, Offset, ?HEADER_SIZE) of
+    case bitcask_file:file_pread(Fd, Offset, ?HEADER_SIZE) of
         {ok, Header} when erlang:size(Header) =:= ?HEADER_SIZE ->
             <<_Crc32:?CRCSIZEFIELD, Tstamp:?TSTAMPFIELD, KeySz:?KEYSIZEFIELD,
               ValueSz:?VALSIZEFIELD>> = Header,
@@ -332,7 +386,7 @@ fold_keys_loop(Fd, Offset, Fun, Acc0) ->
             PosInfo = {Offset, TotalSz},
             %% NOTE: We are intentionally *not* reading the value blob,
             %%       so we cannot check the checksum.
-            case file:pread(Fd, Offset + ?HEADER_SIZE, KeySz) of
+            case bitcask_file:file_pread(Fd, Offset + ?HEADER_SIZE, KeySz) of
                 {ok, Key} when erlang:size(Key) =:= KeySz ->
                     Acc = Fun(Key, Tstamp, PosInfo, Acc0),
                     fold_keys_loop(Fd, Offset + TotalSz, Fun, Acc);
@@ -355,113 +409,121 @@ fold_keys_loop(Fd, Offset, Fun, Acc0) ->
 
 
 fold_hintfile(State, Fun, Acc) ->
-    case file:open(hintfile_name(State), [read, raw, binary, read_ahead]) of
+    HintFile = hintfile_name(State),
+    case bitcask_file:file_open(HintFile, [readonly]) of
         {ok, HintFd} ->
             try
-                {ok, _} = file:position(HintFd, bof),
-                case file:read(HintFd, ?HINT_RECORD_SZ) of
+                {ok, DataI} = file:read_file_info(State#filestate.filename),
+                DataSize = DataI#file_info.size,
+                case bitcask_file:file_read(HintFd, ?HINT_RECORD_SZ) of
                     {ok, <<H:?HINT_RECORD_SZ/bytes>>} ->
-                        fold_hintfile_loop(HintFd, H, Fun, Acc);
+                        fold_hintfile_loop(DataSize, HintFile,
+                                           HintFd, 0, H, Fun, Acc);
+                    {ok, Bytes} ->
+                        error_logger:error_msg("~s:fold_hintfile: ~s: expected "
+                                               "~p bytes but got only ~p "
+                                               "bytes, skipping\n",
+                                   [?MODULE, HintFile,
+                                    ?HINT_RECORD_SZ, size(Bytes)]),
+                        {error, {incomplete_hint, 1}};
                     eof ->
-                        Acc;
+                        {error, empty_hintfile}; % shoudld never be empty hintfiles
                     {error, Reason} ->
                         {error, {fold_hintfile, Reason}}
                 end
             after
-                file:close(HintFd)
+                bitcask_file:file_close(HintFd)
             end;
         {error, Reason} ->
             {error, {fold_hintfile, Reason}}
     end.
 
-fold_hintfile_loop(Fd, HintRecord, Fun, Acc0) ->
+
+fold_hintfile_loop(_DataSize, HintFile, _Fd, HintCRC,
+                   <<0:?TSTAMPFIELD, 0:?KEYSIZEFIELD,
+                     ExpectCRC:?TOTALSIZEFIELD, (?MAXOFFSET):?OFFSETFIELD>>, _Fun, Acc0) ->
+    case HintCRC of
+        ExpectCRC ->
+            Acc0;
+        _ ->
+            error_logger:error_msg("Hintfile '~s' has bad CRC ~p expected ~p\n",
+                                   [HintFile, HintCRC, ExpectCRC]),
+            {error, {bad_crc, HintCRC, ExpectCRC}}
+    end;
+fold_hintfile_loop(DataSize, HintFile, Fd, HintCRC0,
+                   HintRecord, Fun, Acc0) ->
     <<Tstamp:?TSTAMPFIELD, KeySz:?KEYSIZEFIELD,
       TotalSz:?TOTALSIZEFIELD, Offset:?OFFSETFIELD>> = HintRecord,
     ReadSz = KeySz + ?HINT_RECORD_SZ,
-    case file:read(Fd, ReadSz) of
-        {ok, <<Key:KeySz/bytes, Rest/binary>>} ->
+    case bitcask_file:file_read(Fd, ReadSz) of
+        {ok, <<Key:KeySz/bytes, Rest/binary>>} when
+                                               Offset + TotalSz =< DataSize ->
             PosInfo = {Offset, TotalSz},
             Acc = Fun(Key, Tstamp, PosInfo, Acc0),
             case Rest of
                 <<NextRecord:?HINT_RECORD_SZ/bytes>> ->
-                    fold_hintfile_loop(Fd, NextRecord, Fun, Acc);
+                    HintCRC=erlang:crc32(HintCRC0, [HintRecord, Key]),
+                    fold_hintfile_loop(DataSize, HintFile,  Fd, 
+                                       HintCRC, NextRecord, Fun, Acc);
                 <<>> ->
-                    Acc;
+                    %% Hint files without CRCs will end on a record boundary.
+                    %% No way to know whether to expect a crc or not.
+                    %% Over time, merges will add CRCs to all hint files and
+                    %% we can set this as the default.
+                    case application:get_env(bitcask, require_hint_crc) of
+                        {ok, true} ->
+                            {error, {incomplete_hint, 4}};
+                        _ ->
+                            <<>>
+                    end;
                 X ->
                     error_logger:error_msg("Bad hintfile data 1: ~p\n", [X]),
-                    Acc
+                    {error, {incomplete_hint, 2}}
             end;
+        {ok, _} ->
+            error_logger:error_msg("Hintfile '~s' contains pointer ~p ~p "
+                                   "that is greater than total data size ~p\n",
+                                   [HintFile, Offset, TotalSz, DataSize]),
+            {error, {incomplete_hint, 3}};
         eof ->
             {error, incomplete_key};
         {error, Reason} ->
-            {error, Reason};
-        X ->
-            error_logger:error_msg("Bad hintfile data 2: ~p\n", [X]),
-            Acc0
+            {error, Reason}
     end.
 
 create_file_loop(DirName, Opts, Tstamp) ->
     Filename = mk_filename(DirName, Tstamp),
     ok = filelib:ensure_dir(Filename),
-    case bitcask_nifs:create_file(Filename) of
-        true ->
-            {ok, FD} = file:open(Filename, [read, write, raw, binary, read_ahead]),
-            {ok, HintFD} = file:open(hintfile_name(Filename),
-                                     [read, write, raw, binary, read_ahead]),
-            %% If o_sync is specified in the options, try to set that
-            %% flag on the underlying file descriptor
-            case bitcask:get_opt(sync_strategy, Opts) of
-                o_sync ->
-                    %% Make a hacky assumption here that if we open a
-                    %% raw file, we get back a specific tuple from the
-                    %% Erlang VM. The tradeoff is that we can set the
-                    %% O_SYNC flag on the fd, thus improving
-                    %% performance rather dramatically.
-                    {file_descriptor, prim_file, {_Port, RealFd}} = FD,
-                    case bitcask_nifs:set_osync(RealFd) of
-                        ok ->
-                            {ok, #filestate{mode = read_write,
-                                            filename = Filename,
-                                            tstamp = file_tstamp(Filename),
-                                            hintfd = HintFD, fd = FD, ofs = 0}};
-                        {error, Reason} ->
-                            {error, Reason}
-                    end;
-                _ ->
-                    {ok, #filestate{mode = read_write,
-                                    filename = Filename,
-                                    tstamp = file_tstamp(Filename),
-                                    hintfd = HintFD, fd = FD, ofs = 0}}
-            end;
-        false ->
+
+    %% Check for o_sync strategy and add to opts
+    FinalOpts = case bitcask:get_opt(sync_strategy, Opts) of
+                    o_sync ->
+                        [o_sync | Opts];
+                    _ ->
+                        Opts
+                end,
+
+    case bitcask_file:file_open(Filename, FinalOpts) of
+        {ok, FD} ->
+            {ok, HintFD} = bitcask_file:file_open(hintfile_name(Filename), FinalOpts),
+            {ok, #filestate{mode = read_write,
+                            filename = Filename,
+                            tstamp = file_tstamp(Filename),
+                            hintfd = HintFD, fd = FD, ofs = 0}};
+
+        {error, _} ->
             %% Couldn't create a new file with the requested name,
-            %% increment the tstamp by 1 and try again. Conceptually,
-            %% this introduces some drift into the actual creation
+            %% check for the more recent timestamp and increment by
+            %% 1 and try again.
+            %% This introduces some drift into the actual creation
             %% time, but given that we only have at most 2 writers
             %% (writer + merger) for a given bitcask, it shouldn't be
             %% more than a few seconds. The alternative it to sleep
             %% until the next second rolls around -- but this
             %% introduces lengthy, unnecessary delays.
-            create_file_loop(DirName, Opts, Tstamp + 1)
-    end.
-
-generate_hintfile(Filename, {FolderMod, FolderFn, FolderArgs}) ->
-    %% Create the temp file that we will write records out to.
-    TmpFilename = temp_filename(Filename ++ ".~w"),
-    {ok, Fd} = file:open(TmpFilename, [read, write, raw, binary, read_ahead]),
-
-    %% Run the provided fold function over whatever the dataset is. The function
-    %% is passed the Fd as the accumulator argument, and must return the same
-    %% Fd on success. Otherwise, we expect an {error, Reason}.
-    try apply(FolderMod, FolderFn, FolderArgs ++ [Fd]) of
-        Fd ->
-            %% All done writing -- rename to the actual hintfile
-            ok = file:rename(TmpFilename, Filename);
-        {error, Reason} ->
-            {error, Reason}
-    after
-        file:delete(TmpFilename),
-        file:close(Fd)
+            %% This also handles the possibility of clock skew
+            MostRecentTS = most_recent_tstamp(DirName),
+            create_file_loop(DirName, Opts, MostRecentTS + 1)
     end.
 
 hintfile_entry(Key, Tstamp, {Offset, TotalSz}) ->
@@ -469,58 +531,16 @@ hintfile_entry(Key, Tstamp, {Offset, TotalSz}) ->
     [<<Tstamp:?TSTAMPFIELD>>, <<KeySz:?KEYSIZEFIELD>>,
      <<TotalSz:?TOTALSIZEFIELD>>, <<Offset:?OFFSETFIELD>>, Key].
 
-temp_filename(Template) ->
-    temp_filename(Template, now()).
+%% Visitor function for hintfile validation - broken out as a
+%% separate function to make tracing with dbg easier if there is
+%% ever an issue with hintfiles.
+has_valid_hintfile_visitor(_Key, _Tstamp, _PosInfo, Acc0) when is_number(Acc0) ->
+    Acc0+1;
+has_valid_hintfile_visitor(_Key, _Tstamp, _PosInfo, Acc0) ->
+    Acc0.
 
-temp_filename(Template, Seed0) ->
-    {Int, Seed} = random:uniform_s(65536, Seed0),
-    Filename = ?FMT(Template, [Int]),
-    case bitcask_nifs:create_file(Filename) of
-        true ->
-            Filename;
-        false ->
-            temp_filename(Template, Seed)
-    end.
-
-%% ===================================================================
-%% EUnit tests
-%% ===================================================================
--ifdef(TEST).
-
--ifdef(EQC).
-
--define(QC_OUT(P), eqc:on_output(fun(Str, Args) -> ?debugFmt(Str, Args) end, P)).
-
-init_file(Dirname, Kvs) ->
-    os:cmd(?FMT("rm -rf ~s", [Dirname])),
-    {ok, F} = create_file(Dirname, []),
-    {ok, _} = write_kvs(Kvs, F, 0).
-
-write_kvs([], F, _Tstamp) ->
-    {ok, F};
-write_kvs([{K, V} | Rest], F, Tstamp) ->
-    {ok, F2, _, _} = write(F, K, V, Tstamp),
-    write_kvs(Rest, F2, Tstamp + 1).
-
-key_value() ->
-    {eqc_gen:non_empty(binary()), binary()}.
-
-prop_hintfile() ->
-    ?FORALL(Kvs, eqc_gen:non_empty(list(key_value())),
-            begin
-                 {ok, F} = init_file("/tmp/bc.fop.hintfile", Kvs),
-                 ok = create_hintfile(F),
-                 true = has_hintfile(F),
-                 AccFn = fun(K, Tstamp, Pos, Acc0) ->
-                                 [{K, Tstamp, Pos} | Acc0]
-                         end,
-                 fold_keys(F, AccFn, [], datafile) ==
-                     fold_keys(F, AccFn, [], hintfile)
-            end).
-
-prop_hintfile_test_() ->
-    {timeout, 60, fun() -> ?assert(eqc:quickcheck(prop_hintfile())) end}.
-
--endif.
-
--endif.
+%% Find the most recent timestamp for a .data file
+most_recent_tstamp(DirName) ->
+    lists:foldl(fun({TS,_},Acc) -> 
+                       erlang:max(TS, Acc)
+               end, 0, data_file_tstamps(DirName)).
